use std::cmp::{max, min};
use std::collections::HashMap;
use std::time::Instant;

use crate::Part::{Gear, Other};

type CharMatrix = Box<[Box<[char]>]>;

#[derive(Eq, PartialEq, Hash, Clone)]
enum Part {
    Gear,
    Other,
}

#[derive(Eq, PartialEq, Hash, Clone)]
struct PartLocation {
    part: Part,
    row: usize,
    col: usize,
}

fn main() {
    let input = include_str!("../../../input/day03.txt");
    let start = Instant::now();
    let (parts_sum, gear_ratios_sum) = schematic(input);

    println!(
        "part1: {}\npart2: {}\nboth took {} µs",
        parts_sum,
        gear_ratios_sum,
        start.elapsed().as_micros()
    );
}

fn schematic(input: &str) -> (u32, u32) {
    let parsed = to_matrix(&input);

    let mut part_number_sum = 0;
    let mut part_adj_numbers: HashMap<PartLocation, Vec<u32>> = HashMap::new();

    for (line, line_chars) in parsed.iter().enumerate() {
        let mut last_start = 0;
        let mut num = 0;
        for (pos, &char) in line_chars.iter().enumerate() {
            if !char.is_digit(10) {
                if num != 0 {
                    if let Some(parts) = get_adjacent_parts(&parsed, line, last_start, pos - 1) {
                        insert_part_adj_numbers(&mut part_adj_numbers, &parts, num);
                        part_number_sum += num;
                    }
                }

                num = 0;
                continue;
            }

            if num == 0 {
                last_start = pos;
            }

            num = num * 10 + char.to_digit(10).unwrap();
        }

        if num != 0 {
            if let Some(parts) =
                get_adjacent_parts(&parsed, line, last_start, parsed.get(0).unwrap().len() - 1)
            {
                insert_part_adj_numbers(&mut part_adj_numbers, &parts, num);
                part_number_sum += num;
            }
        }
    }

    let gear_ratio_sum = part_adj_numbers
        .iter()
        .filter(|&(p, n)| p.part == Gear && n.len() == 2)
        .map(|(_, n)| n.get(0).unwrap() * n.get(1).unwrap())
        .sum();

    (part_number_sum, gear_ratio_sum)
}

fn insert_part_adj_numbers(
    map: &mut HashMap<PartLocation, Vec<u32>>,
    parts: &Vec<PartLocation>,
    num: u32,
) {
    for part in parts {
        map.entry(part.clone()).or_insert(vec![]).push(num)
    }
}

fn to_matrix(input: &str) -> CharMatrix {
    input
        .lines()
        .into_iter()
        .map(|line| line.chars().into_iter().collect())
        .collect()
}

fn get_adjacent_parts(
    matrix: &CharMatrix,
    line: usize,
    start: usize,
    end: usize,
) -> Option<Vec<PartLocation>> {
    let height = matrix.len();
    let width = matrix.get(0).unwrap().len();

    let start_row = max(line as i32 - 1, 0) as usize;
    let end_row = min(line as i32 + 1, height as i32 - 1) as usize;

    let start_col = max(start as i32 - 1, 0) as usize;
    let end_col = min(end as i32 + 1, width as i32 - 1) as usize;

    let mut parts = vec![];

    for row in start_row..=end_row {
        for col in start_col..=end_col {
            let opt_part = match *(matrix.get(row).unwrap().get(col).unwrap()) {
                '.' | '0'..='9' => None,
                '*' => Some(PartLocation {
                    part: Gear,
                    row,
                    col,
                }),
                _ => Some(PartLocation {
                    part: Other,
                    row,
                    col,
                }),
            };

            if let Some(part) = opt_part {
                parts.push(part);
            }
        }
    }

    if parts.is_empty() {
        None
    } else {
        Some(parts)
    }
}

#[test]
fn test_example_schematic() {
    do_test(
        "467..114..
...*......
..35..633.
......#...
617*......
.....+.58.
..592.....
......755.
...$.*....
.664.598..",
        (4361, 467835),
    )
}

#[test]
fn test_multiple_gears() {
    do_test(
        "2..
.*.
..3
.*.
4..",
        (2 + 3 + 4, 2 * 3 + 3 * 4),
    )
}

fn do_test(input: &str, results: (u32, u32)) {
    assert_eq!(results, schematic(input));
}
