use std::cmp::min;
use std::ops::Range;
use std::str::FromStr;
use std::time::Instant;

#[derive(Debug)]
struct LookupPipeline {
    tables: Vec<LookupTable>,
}

impl LookupPipeline {
    fn map(&self, source: u64) -> u64 {
        let mut val = source;

        for table in &self.tables {
            val = table.map(val);
        }

        val
    }

    fn map_range(&self, source_rng: Range<u64>) -> Vec<Range<u64>> {
        let mut ranges = vec![source_rng];

        for table in &self.tables {
            ranges = table.map_ranges(&ranges);
        }

        ranges
    }
}

impl From<Vec<LookupTable>> for LookupPipeline {
    fn from(value: Vec<LookupTable>) -> Self {
        LookupPipeline { tables: value }
    }
}

#[derive(Debug)]
struct LookupTable {
    ranges: Vec<LookupRange>,
}

impl LookupTable {
    fn find_range(&self, start: u64) -> Option<&LookupRange> {
        self.ranges
            .iter()
            .filter(|r| r.source_start <= start && r.count > start - r.source_start)
            .next()
    }

    fn map(&self, source: u64) -> u64 {
        self.find_range(source)
            .map(|range| range.destination_start + (source - range.source_start))
            .unwrap_or(source) // if there is no mapping, the result is the source itself
    }

    fn map_ranges(&self, source_ranges: &Vec<Range<u64>>) -> Vec<Range<u64>> {
        let mut ranges = vec![];

        for range in source_ranges {
            let mut range_left = range.clone();

            while !range_left.is_empty() {
                let lookup_range = self.find_range(range_left.start);

                match lookup_range {
                    None => {
                        // start is not in a range, find next range
                        let next_start = self
                            .ranges
                            .iter()
                            .skip_while(|r| r.source_start < range_left.start)
                            .next()
                            .map(|rng| rng.source_start);

                        match next_start {
                            None => {
                                ranges.push(range_left.clone());
                                range_left = range_left.end..range_left.end;
                            }
                            Some(next_start) => {
                                let take_len = min(
                                    range_left.end - range_left.start,
                                    next_start - range_left.start,
                                );

                                ranges.push(range_left.start..range_left.start + take_len);
                                range_left = range_left.start + take_len..range_left.end;
                            }
                        }
                    }
                    Some(rng) => {
                        let source_offset = range_left.start - rng.source_start;
                        let take_len =
                            min(range_left.end - range_left.start, rng.count - source_offset);

                        ranges.push(
                            rng.destination_start + source_offset
                                ..rng.destination_start + source_offset + take_len,
                        );
                        range_left = range_left.start + take_len..range_left.end;
                    }
                }
            }
        }

        ranges
    }
}

impl FromStr for LookupTable {
    type Err = ();

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut lines_it = s.lines();
        let mut src_dest_it = lines_it
            .next()
            .unwrap()
            .split(' ')
            .next()
            .unwrap()
            .split("-to-");
        let _source = src_dest_it.next().unwrap();
        let _dest = src_dest_it.next().unwrap();

        let mut ranges: Vec<LookupRange> = lines_it
            .map(|line| line.parse::<LookupRange>().unwrap())
            .collect();
        ranges.sort_unstable_by_key(|r| r.source_start);

        Ok(LookupTable { ranges })
    }
}

#[derive(Debug)]
struct LookupRange {
    source_start: u64,
    destination_start: u64,
    count: u64,
}

impl FromStr for LookupRange {
    type Err = ();

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let nums = parse_numbers(s);
        assert_eq!(nums.len(), 3);

        Ok(LookupRange {
            destination_start: nums[0],
            source_start: nums[1],
            count: nums[2],
        })
    }
}

fn main() {
    let input = include_str!("../../../input/day05.txt");
    let mut block_it = input.split("\r\n\r\n");

    let mut start = Instant::now();
    let seeds = parse_numbers(&block_it.next().unwrap()[7..]); // remove seeds: prefix
    let lookup: LookupPipeline = block_it
        .map(|block| block.parse::<LookupTable>().unwrap())
        .collect::<Vec<_>>()
        .into();
    println!("Parsing took {} µs", start.elapsed().as_micros());

    start = Instant::now();
    let min_location = seeds.iter().map(|s| lookup.map(*s)).min().unwrap();
    println!(
        "part1: {} ({} µs)",
        min_location,
        start.elapsed().as_micros()
    );

    start = Instant::now();
    let min_location_giga = seeds
        .chunks(2)
        .map(|seed_range| seed_range[0]..(seed_range[0] + seed_range[1]))
        .flat_map(|seed_range| lookup.map_range(seed_range))
        .map(|loc_range| loc_range.start)
        .min()
        .unwrap();
    println!(
        "part2: {} ({} µs)",
        min_location_giga,
        start.elapsed().as_micros()
    );
}

fn parse_numbers(line: &str) -> Vec<u64> {
    line.split_whitespace()
        .filter(|s| !s.trim().is_empty())
        .map(|num_str| num_str.parse::<u64>().unwrap())
        .collect()
}

#[test]
fn test_map_multiple_ranges() {
    let table1 = LookupTable {
        ranges: vec![
            LookupRange {
                source_start: 5,
                destination_start: 10,
                count: 5,
            },
            LookupRange {
                source_start: 10,
                destination_start: 30,
                count: 1,
            },
        ],
    };

    let seed_range = 0..20u64;
    let res1 = table1.map_ranges(&vec![seed_range]);

    assert_eq!(vec![0..5, 10..15, 30..31, 11..20], res1);
}

#[test]
fn test_map_single_fully_overlapping_range() {
    let table = LookupTable {
        ranges: vec![LookupRange {
            source_start: 0,
            destination_start: 20,
            count: 10,
        }],
    };

    let seed_range = 0..10;
    let res = table.map_ranges(&vec![seed_range]);

    assert_eq!(vec![20..30], res);
}

#[test]
fn test_map_no_range() {
    let table = LookupTable { ranges: vec![] };

    let seed_range = 0..10;
    let res = table.map_ranges(&vec![seed_range]);

    assert_eq!(vec![0..10], res);
}

#[test]
fn test_multiple_seeds_no_range() {
    let table = LookupTable { ranges: vec![] };

    let seed_ranges = vec![0..10, 50..100];
    let res = table.map_ranges(&seed_ranges);

    assert_eq!(vec![0..10, 50..100], res);
}

#[test]
fn test_multiple_tables() {
    let table1 = LookupTable {
        ranges: vec![LookupRange {
            source_start: 10,
            destination_start: 20,
            count: 5,
        }],
    };
    let table2 = LookupTable {
        ranges: vec![LookupRange {
            source_start: 22,
            destination_start: 15,
            count: 5,
        }],
    };
    let table3 = LookupTable {
        ranges: vec![LookupRange {
            source_start: 13,
            destination_start: 10,
            count: 5,
        }],
    };
    let lookup: LookupPipeline = vec![table1, table2, table3].into();

    let mut res = lookup.map_range(0..15);
    res.sort_unstable_by_key(|r| r.start);
    assert_eq!(vec![0..10, 12..15, 20..22], res);
}
