use std::time::Instant;

fn main() {
    let input = include_str!("../../../input/day06.txt");

    let mut start = Instant::now();
    let time_distance_pairs = parse(input);
    println!("Parsing took {} µs", start.elapsed().as_micros());

    let res1b = solve(&time_distance_pairs);
    println!("part1: {} ({} µs)", res1b, start.elapsed().as_micros());

    let combined = combine(&time_distance_pairs);
    start = Instant::now();
    let res2b = solve(&vec![combined]);
    println!("part2: {} ({} µs)", res2b, start.elapsed().as_micros());
}

fn solve(pairs: &Vec<(u64, u64)>) -> u64 {
    let mut acc = 1;

    for &(time, dist) in pairs {
        //      hold * (time - hold) > dist in \N
        // <=>  hold * time - hold^2 - dist = 0
        // <=> -hold^2 + time * hold - dist
        // a = -1,   b = time,    c = -dist

        // discriminant b^2 - 4ac
        let d_sqrt = ((time.pow(2) - 4 * dist) as f64).sqrt();
        // no need to check for < 1, there are no such cases in the data

        let max = f64::floor(((time as f64) + d_sqrt) / 2f64);
        let min = f64::ceil(((time as f64) - d_sqrt) / 2f64);

        acc *= (max - min + 1.) as u64;
    }

    acc
}

fn parse(input: &str) -> Vec<(u64, u64)> {
    let mut lines_it = input.lines();

    let times = read_numbers(lines_it.next().unwrap());
    let distances = read_numbers(lines_it.next().unwrap());

    times.into_iter().zip(distances).collect()
}

fn read_numbers(line: &str) -> Vec<u64> {
    line.split(char::is_whitespace)
        .map(|part| part.trim())
        .filter(|part| !part.is_empty() && part.chars().next().unwrap().is_numeric())
        .map(|num_str| num_str.parse().unwrap())
        .collect()
}

fn combine(pairs: &Vec<(u64, u64)>) -> (u64, u64) {
    let mut t_acc = 0;
    let mut d_acc = 0;

    let mut pair_it = pairs.iter().peekable();
    while let Some(&(t, d)) = pair_it.next() {
        t_acc += t;
        d_acc += d;

        if let Some((next_t, next_d)) = pair_it.peek() {
            t_acc *= 10u64.pow(u64::ilog10(*next_t) + 1);
            d_acc *= 10u64.pow(u64::ilog10(*next_d) + 1);
        }
    }

    (t_acc, d_acc)
}
