use std::cmp::Ordering;
use std::collections::HashMap;
use std::str::FromStr;
use std::time::Instant;

use crate::HandKind::{
    FiveOfAKind, FourOfAKind, FullHouse, HighCard, OnePair, ThreeOfAKind, TwoPair,
};

type CardValues = [char; 13];

const CARD_ORDER_PART_1: CardValues = [
    'A', 'K', 'Q', 'J', 'T', '9', '8', '7', '6', '5', '4', '3', '2',
];
const CARD_ORDER_PART_2: CardValues = [
    'A', 'K', 'Q', 'T', '9', '8', '7', '6', '5', '4', '3', '2', 'J',
];

#[derive(PartialOrd, Ord, PartialEq, Eq, Debug)]
enum HandKind {
    HighCard, // lowest value
    OnePair,
    TwoPair,
    ThreeOfAKind,
    FullHouse,
    FourOfAKind,
    FiveOfAKind,
}

#[derive(Debug)]
struct Hand {
    cards: [char; 5],
    bid: u64,
}

impl FromStr for Hand {
    type Err = ();

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut it = s.split(" ");
        let card_str = it.next().unwrap();
        let cards = card_str.chars().collect::<Vec<_>>().try_into().unwrap();
        let bid = it.next().unwrap().parse::<u64>().unwrap();

        Ok(Hand { cards, bid })
    }
}

#[derive(Debug)]
struct ValuedHand<'a> {
    original_hand: &'a Hand,
    card_values: [u8; 5],
    kind: HandKind,
}

impl<'a> From<(&'a Hand, &CardValues, Option<char>)> for ValuedHand<'a> {
    fn from(value: (&'a Hand, &CardValues, Option<char>)) -> Self {
        let (hand, values, joker) = value;

        let card_values: [u8; 5] = hand
            .cards
            .iter()
            .map(|&c| (values.len() - values.iter().position(|&oc| oc == c).unwrap()) as u8)
            .collect::<Vec<_>>()
            .try_into()
            .unwrap();

        let mut card_counts: HashMap<&u8, u8> = HashMap::new();
        card_values
            .iter()
            .for_each(|c| *card_counts.entry(c).or_insert(0) += 1);

        if let Some(joker_card) = joker {
            let joker_value =
                (values.len() - values.iter().position(|&oc| oc == joker_card).unwrap()) as u8;

            // skip case where no jokers or all cards are joker
            if card_counts.len() != 1 && card_values.contains(&joker_value) {
                let joker_count = card_counts.remove(&joker_value).unwrap();

                let (&next_most_card, _) = card_counts
                    .iter()
                    // order by count desc
                    .max_by(|&(_, ac), &(_, bc)| ac.cmp(bc))
                    .unwrap();

                *(card_counts
                    .entry(next_most_card)
                    .or_insert_with(|| panic!())) += joker_count;
            }
        }

        let mut counts: Vec<u8> = card_counts.into_values().collect();
        counts.sort_unstable_by(|a, b| b.cmp(a));

        let kind = match &counts[..] {
            [5] => FiveOfAKind,
            [4, _] => FourOfAKind,
            [3, 2] => FullHouse,
            [3, 1, _] => ThreeOfAKind,
            [2, 2, _] => TwoPair,
            [2, 1, _, _] => OnePair,
            [1, _, _, _, _] => HighCard,
            _ => unreachable!(),
        };

        ValuedHand {
            original_hand: hand,
            card_values,
            kind,
        }
    }
}

impl Eq for ValuedHand<'_> {}

impl PartialEq<Self> for ValuedHand<'_> {
    fn eq(&self, other: &Self) -> bool {
        self.kind == other.kind && self.card_values == other.card_values
    }
}

impl PartialOrd<Self> for ValuedHand<'_> {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        let kind_order = self.kind.cmp(&other.kind);

        Some(if kind_order != Ordering::Equal {
            kind_order
        } else {
            self.card_values.cmp(&other.card_values)
        })
    }
}

impl Ord for ValuedHand<'_> {
    fn cmp(&self, other: &Self) -> Ordering {
        self.partial_cmp(other).unwrap()
    }
}

fn main() {
    let input = include_str!("../../../input/day07.txt");
    let mut start = Instant::now();
    let hands: Vec<Hand> = input
        .lines()
        .map(|line| line.parse::<Hand>().unwrap())
        .collect();
    println!("Parsing took {} µs", start.elapsed().as_micros());

    start = Instant::now();
    let mut part1_valued_hands = hands
        .iter()
        .map(|h| ValuedHand::from((h, &CARD_ORDER_PART_1, None)))
        .collect::<Vec<_>>();
    part1_valued_hands.sort_unstable();
    let res1: u64 = part1_valued_hands
        .iter()
        .enumerate()
        .map(|(i, h)| (i as u64 + 1) * h.original_hand.bid)
        .sum();
    println!("part1: {} ({} µs)", res1, start.elapsed().as_micros());

    start = Instant::now();
    let mut part2_valued_hands = hands
        .iter()
        .map(|h| ValuedHand::from((h, &CARD_ORDER_PART_2, Some('J'))))
        .collect::<Vec<_>>();
    part2_valued_hands.sort_unstable();
    let res2: u64 = part2_valued_hands
        .iter()
        .enumerate()
        .map(|(i, h)| (i as u64 + 1) * h.original_hand.bid)
        .sum();
    println!("part2: {} ({} µs)", res2, start.elapsed().as_micros());
}

#[cfg(test)]
mod test {
    use crate::*;

    fn test_part2(cards: &str, expected_values: &[u8; 5], expected_kind: HandKind) {
        let hand = format!("{} 0", cards).parse::<Hand>().unwrap();
        let valued_hand = ValuedHand::from((&hand, &CARD_ORDER_PART_2, Some('J')));

        assert_eq!(expected_values, &valued_hand.card_values);
        assert_eq!(expected_kind, valued_hand.kind);
    }

    #[test]
    fn joker_replacement_five() {
        test_part2("AAAAJ", &[13, 13, 13, 13, 1], FiveOfAKind);
    }

    #[test]
    fn joker_replacement_four() {
        test_part2("AAAKJ", &[13, 13, 13, 12, 1], FourOfAKind);
    }
}
