use rayon::prelude::*;
use std::collections::HashMap;
use std::time::Instant;

type Nodes<'a> = HashMap<&'a str, (&'a str, &'a str)>;

fn main() {
    let input = include_str!("../../../input/day08.txt");
    let mut start = Instant::now();
    let (instructions, str_nodes) = parse(input);
    println!("Parsing took {} µs", start.elapsed().as_micros());

    start = Instant::now();
    let res1 = part1_steps(&instructions, &str_nodes);
    println!("part1: {} ({} µs)", res1, start.elapsed().as_micros());

    start = Instant::now();
    let res2 = part2_steps(&instructions, &str_nodes);
    println!("part2: {} ({} µs)", res2, start.elapsed().as_micros());
}

fn parse(input: &str) -> (Vec<char>, Nodes) {
    let mut lines_it = input.lines();
    let instructions = lines_it.next().unwrap().chars().collect::<Vec<_>>();
    lines_it.next(); // skip empty line

    let str_nodes = lines_it
        .map(|line| {
            let mut name_groups_it = line.split('=');
            let name = name_groups_it.next().unwrap().trim();
            let mut groups_it = name_groups_it.next().unwrap().split(",");
            let tr: &[_] = &[' ', '(', ')'];
            let left = groups_it.next().unwrap().trim_matches(tr);
            let right = groups_it.next().unwrap().trim_matches(tr);

            (name, (left, right))
        })
        .collect();

    (instructions, str_nodes)
}

fn part1_steps(instructions: &Vec<char>, nodes: &Nodes) -> u32 {
    let mut current_node = "AAA";
    let mut instr_idx = 0usize;
    let mut count = 0u32;

    while current_node != "ZZZ" {
        let instr = instructions[instr_idx];
        current_node = next_node(current_node, nodes, &instr);

        count += 1;
        instr_idx += 1;
        if instr_idx >= instructions.len() {
            instr_idx = 0;
        }
    }

    count
}

fn part2_steps(instructions: &Vec<char>, nodes: &Nodes) -> u64 {
    let cycle_lengths = nodes
        .par_iter()
        .filter(|&(k, _)| k.ends_with('A'))
        .map(|(k, _)| *k)
        .map(|start| cycle_length(instructions, nodes, start) as u64)
        .collect::<Vec<_>>();

    lcm(&cycle_lengths)
}

fn cycle_length(instructions: &Vec<char>, nodes: &Nodes, start: &str) -> u32 {
    let mut count = 0u32;
    let mut last_target: Option<u32> = None;
    let mut current_node = start;
    let mut instruction_idx = 0usize;

    loop {
        let instruction = instructions[instruction_idx];
        current_node = next_node(current_node, nodes, &instruction);

        count += 1;
        instruction_idx = (instruction_idx + 1) % instructions.len();

        if let Some(cl) = match current_node.ends_with('Z') {
            true => match last_target {
                None => {
                    last_target = Some(count);
                    None
                }
                Some(l) => Some(count - l),
            },
            _ => None,
        } {
            return cl;
        }
    }
}

fn lcm(numbers: &[u64]) -> u64 {
    match numbers {
        &[a] => a,
        [a, rest @ ..] => {
            let b = lcm(&rest);
            *a * b / gcd(*a, b)
        }
        [] => 1,
    }
}

fn gcd(a: u64, b: u64) -> u64 {
    match (a, b) {
        (a, 0) => a,
        (a, b) => gcd(b, a % b),
    }
}

#[inline(always)]
fn next_node<'a>(
    curr_node: &str,
    nodes: &HashMap<&str, (&'a str, &'a str)>,
    instruction: &char,
) -> &'a str {
    let node = nodes.get(curr_node).unwrap();

    match instruction {
        'L' => node.0,
        'R' => node.1,
        _ => unreachable!(),
    }
}
