use crate::Tile::{
    EastSouth, Ground, Horizontal, NorthEast, NorthWest, SouthWest, Start, Vertical,
};
use colored::Colorize;
use std::fmt::{Display, Formatter};
use std::time::Instant;

type Grid = Vec<Vec<&'static Tile>>;
type IOGrid = Vec<Vec<IOTile>>;

#[derive(Eq, PartialEq, Debug)]
enum Tile {
    Vertical,
    Horizontal,
    NorthEast,
    EastSouth,
    SouthWest,
    NorthWest,
    Ground,
    Start,
}

impl Display for Tile {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Vertical => "┃",
                Horizontal => "━",
                NorthEast => "┗",
                EastSouth => "┏",
                SouthWest => "┓",
                NorthWest => "┛",
                Ground => " ",
                Start => "╳",
            }
        )
    }
}

#[derive(PartialEq, Eq, Clone)]
enum IOTile {
    FromDown(&'static Tile),
    TargetDown(&'static Tile),
    OtherPipe(&'static Tile),
    Ground,
}

const DEBUG: bool = true;

fn main() {
    let input = include_str!("../../../input/day10.txt");

    let mut start = Instant::now();
    let grid = parse(input);
    let t1 = start.elapsed().as_micros();
    println!(
        "Parsing {}x{} grid took {} µs",
        grid.len(),
        grid[0].len(),
        t1
    );

    start = Instant::now();
    let start_pos = find_start(&grid);
    let (len, io_grid) = follow_pipe_length(&grid, start_pos, 0, None);
    let t2 = start.elapsed().as_micros();
    println!("Pipe following took {} µs", t2);

    start = Instant::now();
    let res1 = len / 2;
    let t3 = start.elapsed().as_micros();
    println!("part1: {} ({} µs)", res1, t3);

    start = Instant::now();
    let res2 = process_io_grid(&io_grid);
    let t4 = start.elapsed().as_micros();
    println!("part2: {} ({} µs)", res2, t4);

    println!("{},{},{},{}", t1, t2, t3, t4);
}

fn process_io_grid(io_grid: &IOGrid) -> u32 {
    let x_size = io_grid.len();
    let y_size = io_grid[0].len();

    let mut line_counter = 0i32;
    let mut in_count = 0;

    for x in 0..x_size {
        for y in 0..y_size {
            let io_tile = &io_grid[x][y];

            match io_tile {
                IOTile::FromDown(_) => line_counter += 1,
                IOTile::TargetDown(_) => line_counter -= 1,
                _ => {}
            }

            let is_in = line_counter != 0;
            if is_in && *io_tile == IOTile::Ground {
                in_count += 1;
            }

            if DEBUG {
                print!(
                    "{}",
                    match *io_tile {
                        IOTile::FromDown(t) | IOTile::TargetDown(t) | IOTile::OtherPipe(t) =>
                            t.to_string(),
                        IOTile::Ground =>
                            if is_in {
                                "I".bright_blue().to_string()
                            } else {
                                "O".bright_red().to_string()
                            },
                    }
                );
            }
        }

        line_counter = 0;

        if DEBUG {
            println!();
        }
    }

    in_count
}
fn parse(input: &str) -> Grid {
    // . -> y
    // |/  [x][y]
    // x

    input
        .lines()
        .map(|l| {
            l.chars()
                .map(|c| match c {
                    '|' => &Vertical,
                    '-' => &Horizontal,
                    'L' => &NorthEast,
                    'F' => &EastSouth,
                    '7' => &SouthWest,
                    'J' => &NorthWest,
                    '.' => &Ground,
                    'S' => &Start,
                    _ => unreachable!(),
                })
                .collect()
        })
        .collect()
}

fn find_start(grid: &Grid) -> (usize, usize) {
    let x_size = grid.len();
    let y_size = grid[0].len();

    let start_pos = (0..x_size)
        .flat_map(|x| (0..y_size).map(move |y| (x, y)))
        .find(|(x, y)| grid[*x][*y] == &Start)
        .unwrap();

    start_pos
}

fn follow_pipe_length(
    grid: &Grid,
    pos: (usize, usize),
    len: u32,
    previous: Option<((usize, usize), &Tile)>,
) -> (u32, IOGrid) {
    let mut pos = pos;
    let mut len = len;
    let mut previous = previous;

    let x_size = grid.len();
    let y_size = grid[0].len();

    let mut io_grid = vec![vec![IOTile::Ground; y_size]; x_size];

    loop {
        let (x, y) = pos;
        let tile: &'static Tile = grid[x][y];

        if DEBUG {
            println!("({:3},{:3}) {:2} n: {} ", pos.0, pos.1, tile, len);
        }

        if previous.is_some() && tile == &Start {
            return (len, io_grid);
        }

        let p = previous.clone();
        let mut next = |nx, ny| {
            pos = (nx, ny);
            len += 1;
            previous = Some(((x, y), tile));
        };

        match p {
            None => {
                // Start, find any connecting pipe
                let north = grid[x - 1][y];
                let east = grid[x][y + 1];
                let south = grid[x + 1][y];
                let west = grid[x][y - 1];

                if *north == Vertical || *north == SouthWest || *north == EastSouth {
                    io_grid[x][y] = IOTile::FromDown(tile);
                    next(x - 1, y)
                } else if *east == Horizontal || *east == NorthWest || *east == SouthWest {
                    io_grid[x][y] = IOTile::OtherPipe(tile);
                    next(x, y + 1)
                } else if *south == Vertical || *south == NorthEast || *south == NorthWest {
                    io_grid[x][y] = IOTile::TargetDown(tile);
                    next(x + 1, y)
                } else if *west == Horizontal || *west == EastSouth || *west == NorthEast {
                    io_grid[x][y] = IOTile::OtherPipe(tile);
                    next(x, y - 1)
                } else {
                    panic!("Could not find any connecting pipe at start");
                }
            }
            Some(((px, py), pt)) => {
                if px < x && py == y {
                    // from north

                    match tile {
                        Vertical => {
                            io_grid[x][y] = IOTile::TargetDown(tile);
                            next(x + 1, y)
                        }
                        NorthEast => {
                            io_grid[x][y] = IOTile::OtherPipe(tile);
                            next(x, y + 1)
                        }
                        NorthWest => {
                            io_grid[x][y] = IOTile::OtherPipe(tile);
                            next(x, y - 1)
                        }
                        _ => panic!(
                            "Illegal pipe connection at ({:3},{:3}) {:2} into ({:3},{:3}) {:2}",
                            px, py, pt, x, y, tile
                        ),
                    }
                } else if px == x && py > y {
                    // from east

                    match tile {
                        Horizontal => {
                            io_grid[x][y] = IOTile::OtherPipe(tile);
                            next(x, y - 1)
                        }
                        EastSouth => {
                            io_grid[x][y] = IOTile::TargetDown(tile);
                            next(x + 1, y)
                        }
                        NorthEast => {
                            io_grid[x][y] = IOTile::OtherPipe(tile);
                            next(x - 1, y)
                        }
                        _ => panic!(
                            "Illegal pipe connection at ({:3},{:3}) {:2} into ({:3},{:3}) {:2}",
                            px, py, pt, x, y, tile
                        ),
                    }
                } else if px > x && py == y {
                    // from south

                    match tile {
                        Vertical => {
                            io_grid[x][y] = IOTile::FromDown(tile);
                            next(x - 1, y)
                        }
                        SouthWest => {
                            io_grid[x][y] = IOTile::FromDown(tile);
                            next(x, y - 1)
                        }
                        EastSouth => {
                            io_grid[x][y] = IOTile::FromDown(tile);
                            next(x, y + 1)
                        }
                        _ => panic!(
                            "Illegal pipe connection at ({:3},{:3}) {:2} into ({:3},{:3}) {:2}",
                            px, py, pt, x, y, tile
                        ),
                    }
                } else if px == x && py < y {
                    // from west

                    match tile {
                        Horizontal => {
                            io_grid[x][y] = IOTile::OtherPipe(tile);
                            next(x, y + 1)
                        }
                        SouthWest => {
                            io_grid[x][y] = IOTile::TargetDown(tile);
                            next(x + 1, y)
                        }
                        NorthWest => {
                            io_grid[x][y] = IOTile::OtherPipe(tile);
                            next(x - 1, y)
                        }
                        _ => panic!(
                            "Illegal pipe connection at ({:3},{:3}) {:2} into ({:3},{:3}) {:2}",
                            px, py, pt, x, y, tile
                        ),
                    }
                } else {
                    panic!("Cannot determine incoming direction")
                }
            }
        }
    }
}
