use std::time::Instant;

const DEBUG: bool = false;

fn main() {
    let input = include_str!("../../../input/day11.txt");

    let mut start = Instant::now();
    let part1_galaxies = parse(input, 2);
    let res1 = calculate_distance_sum(&part1_galaxies);
    let t1 = start.elapsed().as_micros();
    println!("part1: {} ({} µs)", res1, t1);

    start = Instant::now();
    let part2_galaxies = parse(input, 1_000_000);
    let res2 = calculate_distance_sum(&part2_galaxies);
    let t2 = start.elapsed().as_micros();
    println!("part2: {} ({} µs)", res2, t2);

    println!("{},{}", t1, t2);
}

fn parse(input: &str, expansion_factor: u64) -> Vec<(u64, u64)> {
    let lines = input.lines().collect::<Vec<_>>();

    let row_length = lines[0].len();
    let col_length = lines.len();

    let mut empty_rows = vec![true; row_length];
    let mut empty_cols = vec![true; col_length];

    for (row, line) in lines.iter().enumerate() {
        let mut row_empty = true;
        for (col, c) in line.chars().enumerate() {
            if c != '.' {
                empty_cols[col] = false;
                row_empty = false;
            }
        }
        empty_rows[row] = row_empty;
    }

    let empty_row_scan = empty_rows
        .iter()
        .scan(0u64, |s, &empty| match empty {
            true => {
                *s += expansion_factor - 1;
                Some(*s)
            }
            false => Some(*s),
        })
        .collect::<Vec<_>>();
    let empty_col_scan = empty_cols
        .iter()
        .scan(0u64, |s, &empty| match empty {
            true => {
                *s += expansion_factor - 1;
                Some(*s)
            }
            false => Some(*s),
        })
        .collect::<Vec<_>>();

    if DEBUG {
        println!(
            "empty_row_scan: {:?}\nempty_col_scan: {:?}",
            empty_row_scan, empty_col_scan
        );
    }

    let mut galaxies = vec![];

    for (row, line) in input.lines().enumerate() {
        for (col, c) in line.chars().enumerate() {
            if c == '#' {
                galaxies.push((
                    row as u64 + empty_row_scan[row],
                    col as u64 + empty_col_scan[col],
                ));
            }
        }
    }

    galaxies
}

fn calculate_distance_sum(galaxies: &Vec<(u64, u64)>) -> u64 {
    pairs(galaxies).iter().map(|pair| dist(&pair)).sum::<u64>()
}

fn pairs(galaxies: &Vec<(u64, u64)>) -> Vec<((u64, u64), (u64, u64))> {
    let n = galaxies.len();

    (0..n)
        .flat_map(move |s| ((s + 1)..n).map(move |r| (s, r)))
        .map(|(a, b)| (galaxies[a], galaxies[b]))
        .collect()
}

fn dist(pair: &((u64, u64), (u64, u64))) -> u64 {
    let ((x1, y1), (x2, y2)) = pair;

    x1.abs_diff(*x2) as u64 + y1.abs_diff(*y2) as u64
}

#[cfg(test)]
mod test {
    use crate::{dist, pairs, parse};

    #[test]
    fn test_demo() {
        let galaxies = parse(
            "...#......
.......#..
#.........
..........
......#...
.#........
.........#
..........
.......#..
#...#.....",
            2,
        );

        assert_eq!(
            vec![
                (0, 4),
                (1, 9),
                (2, 0),
                (5, 8),
                (6, 1),
                (7, 12),
                (10, 9),
                (11, 0),
                (11, 5)
            ],
            galaxies
        );
    }

    #[test]
    fn test_distances() {
        assert_eq!(6, dist(&((0, 4), (1, 9))));
        assert_eq!(6, dist(&((0, 4), (2, 0))));
        assert_eq!(10, dist(&((1, 9), (2, 0))));
    }

    #[test]
    fn test_pairs_2() {
        assert_eq!(vec![((0, 1), (3, 4))], pairs(&vec![(0, 1), (3, 4)]));
    }

    #[test]
    fn test_pairs_3() {
        assert_eq!(
            vec![((0, 1), (3, 4)), ((0, 1), (5, 10)), ((3, 4), (5, 10))],
            pairs(&vec![(0, 1), (3, 4), (5, 10)])
        );
    }
}
